# -*- coding: utf-8 -*-

"""Fit histogram data coming in on stdin to a user-defined model.  You
should write a python script sublcassing fit-hist.Model."""

import scipy.optimize, scipy
import sys

kB=1.3806503e-23 # m²·kg/s²·K

class Model (object) :

    """For minimizing fit error: returns the normalized rms difference
    between a dataset and a function y(X).  In order to use this
    class, you must sublcass it and overwrite Model.model(x,params).
    It is recommended that you also overwrite
    Model.guess_initial_params(), and then not pass initializing
    params to Model.model().  This helps ensure good automation for
    unsupervised testing.  Developing good heuristics for
    Model.guessInitialParams() is usually the hardest part writing a
    theoretical histogram test.

    For bonus points, you can override Model.printModel() and
    Model.printParams(), to give your users an easy to understand idea
    of what's going on.  It's better to be verbose and clear.
    Remember, this is the testing branch.  It needs to be solid.  Save
    your wonderful, sneaky hacks for the main code ;).

    """
    def __init__(self, numParams, data) :
        """Initialize the Model instance with a set of data, for
        example: data = [(0,4),(1,3),(2,3),(4,0)]"""
        self.data = data # an iterable filled with (x,y) pairs
        # N (# unfolding events) = the sum of the y values (# hits per bin)
        self.N = sum(zip(*data)[1]) # zip(* takes data to [[x0,x1,..],[y..]]
        # assumes constant binwidth
        self.binwidth = self.data[1][0] - self.data[0][0]
        self.numParams = numParams
    def nrms(self, params) :
        """Returns the normalized root mean squagred between the data
        and a model for a given set of parameters"""
        s = 0.0
        for x,y in self.data :
            yResidual = y-self.model(x, params)
            s += yResidual**2
        s /= float(len(self.data))
        return s
    def fit(self, x0=None) :
        """Minimize error given an initial guess vector x0, the heart of
        the class."""
        if x0 == None :
            x0 = self.guessInitialParams()
        xopt = scipy.optimize.fmin_powell(self.nrms, x0)
        return xopt
    def printModel(self) :
        "OVERRIDE: Give the user some idea of what model we're fitting with"
        print "model string not defined"
    def printParams(self) :
        "OVERRIDE: Give the user nicely named format for a parameter vector"
        for i,p in zip(range(len(params)), params) :
            print "param_%d\t=\t%g\n" % (i,p)
    def model(self, x, params) :
        "OVERRIDE: Implement your model"
        assert len(params) == self.numParams
        raise Exception, "model method not defined"
        return 1.0
    def guessInitialParams(self) :
        """OVERRIDE: Provide some intelligent heuristic to pick a
        starting point for minimizing your model"""
        return [1.0] * self.numParams

def readData(fileObj) :
    """Read histogram data from a whitespace-seperated, two-column
    ASCII text-file into a list of tuples"""
    data = []
    for line in fileObj :
        if line[0] == "#" : continue
        data.append([float(x) for x in line.split()])
    return data

def run(numParams, modelClass) :
    import sys
    data = readData(sys.stdin)
    mod = modelClass(numParams, data)
    popt = mod.fit()
    mod.printModel()
    mod.printParams(popt)
    print "N:\t%g" % mod.N
    print "bin:\t%g" % mod.binwidth
